// --- [ TWI_Slave.c ] --------------------------------------------------------
//
//       tab = 3 - 22.02.2012
//
// nach AppNote : AVR311 - TWI Slave Implementation - mit einigen Vernderungen ...

#include <avr/interrupt.h>
#include <avr/io.h>
#include <inttypes.h>
#include "TWI_Slave.h"

#define  TRUE  1
#define  FALSE 0

uint8_t TWI_Tx_buf[SLA_TX_BUF_SIZE];      // Puffer fr Daten, die der Master abruft
uint8_t TWI_Rx_buf[SLA_RX_BUF_SIZE];      // Puffer fr Daten, die vom Master empfangen wurden

volatile uint8_t TWI_Rx_Idx;					// Index/Zhler auf die eingegangenen Bytes
volatile uint8_t TWI_Tx_Idx;              // Index auf die versandten Bytes

uint8_t TWI_msgSize = 0;           	      // Number of bytes to be transmitted.
uint8_t TWI_state   = TWI_NO_STATE;	      // State byte. Default set to TWI_NO_STATE.
uint8_t LastTransOK = FALSE;

uint8_t * Tx_Ptr = TWI_Tx_buf;            // Ein Pointer auf den Start des Arrays mit den zu sendenden Daten
                                          // Per Default ist das das Array TWI_Tx_buf[]

// ----------------------------------------------------------------------------
// Initialisiert den Slave mit einer TWI-Adresse und macht ihn bereit
// zum Empfang von Daten
void TWI_SLA_Init( void )
{
	TWAR  = TWI_SLA_ADR;	                   // Set own TWI slave address.
   TWAMR = TWI_SLA_ADR + 0x06;             // Der Slave lauscht auf den Adressen 8 sowie zustzlich auf 2 und 4

	TWDR = 0xFF;                            // Default content = SDA released.
   TWI_state = TWI_NO_STATE ;
	TWCR = (1<<TWEN)|                       // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag.
          (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Prepare to ACK next time the Slave is addressed.
          (0<<TWWC);
}


// ----------------------------------------------------------------------------
// Bei Bedarf kann mit dieser Funktion der TWI-Status ausgelesen werden
/*
uint8_t TWI_Get_State_Info( void )
{
	while ( TWI_BUSY );                       // Wait until TWI has completed the transmission.
	return ( TWI_state );                     // Return error state.
}
*/

// ----------------------------------------------------------------------------
// Gibt die Anzahl der empfangenen Bytes zurck,
// Rckgabe : 0  - sind keine Daten empfangen worden
// Rckgabe : 1  - es ist genau 1 Byte empfangen - die eigene Adresse + Write-Bit
// Rckgabe : >1 - es sind (n - 1) Datenbyte empfangen, zustzlich die Adresse in Byte.0

uint8_t TWI_SLA_Rx_Cnt(void)
{
if ( TWI_BUSY ) return(0);					   // solange der Empfang aktiviert ist oder eine bertragung luft
return (TWI_Rx_Idx);								// der Pointer beeinhaltet die Anzahl der empf. Bytes
}

// ----------------------------------------------------------------------------
// Stellt Daten im lokalen Array TWI_Tx_buf[] zur Abholung durch den Master bereit
void TWI_SLA_Put_Data( uint8_t *msg, uint8_t msgSize )
{
	uint8_t i;

	TWI_msgSize = msgSize;                       // Number of data to transmit.

	for (i = 0; i < msgSize; i++)                // Copy data that may be transmitted if the TWI Master requests data.
		TWI_Tx_buf[i] = msg[i];
}

//----------------------------------------------------------------------------
// Verndert den Pointer auf die zu versendenden Daten auf die Adresse, die als
// Parameter angegeben ist.
//
void TWI_SLA_Set_TxPtr(uint8_t * data)
{
   Tx_Ptr = data;
}

// ----------------------------------------------------------------------------
// Kopiert die empfangenen Daten im Array TWI_Rx_buf[] in das Array
// dessen Adresse bergeben wurde

uint8_t TWI_SLA_Get_Data( uint8_t *msg)
{
	uint8_t i;

	while ( TWI_BUSY );             		         // Wait until TWI is ready for next transmission.

	if( LastTransOK )               					// Last transmission completed successfully.
	{                                            // Copy data from Transceiver buffer.
  		for (i = 0; i < TWI_Rx_Idx; i++) msg[i] = TWI_Rx_buf[i];
	}
	TWI_Rx_Idx = 0;									   // damit die Daten nur einmal gelesen werden

	TWI_state  = TWI_NO_STATE ;                  // Den Empfang wieder starten
	TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|      // Prepare to ACK next time the Slave is addressed.
          (0<<TWWC);

	return( LastTransOK );
}

// ----------------------------------------------------------------------------
// Die Interruptroutine vollbringt die eigentliche Arbeit ...
ISR(TWI_vect)
{
	switch (TWSR)
	{
	// --- TWI Slave Transmitter status codes ----------------------------------

   case TWI_STX_ADR_ACK:            				// Own SLA+R has been received; ACK has been returned
// case TWI_STX_ADR_ACK_M_ARB_LOST: 				// Arbitration lost in SLA+R/W as Master; own SLA+R has been received;
																// ACK has been returned

      TWI_Tx_Idx = 0;                        	// Set buffer pointer to first data location

   case TWI_STX_DATA_ACK:           				// Data byte in TWDR has been transmitted; ACK has been received
																// Eine berprfung wird eingebaut:
																// der Pointer wird nicht mehr incrementiert, wenn TWI_BUFFER_SIZE erreicht ist
		if (TWI_Tx_Idx < SLA_TX_BUF_SIZE)
		{

         TWDR = Tx_Ptr[TWI_Tx_Idx++];           // Ein Byte in das Sendereister schreiben
//		   TWDR = TWI_Tx_buf[TWI_Tx_Idx++];       // so gehts ohne Pointer
      }
                                                // Es werden solange Daten gesendet, bis der Master mit NACK beendet
																// damit kann der Pointer ber die Arraygrenze hinauslaufen ... your problem

      TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|              // Enable TWI Interupt and clear the flag to send byte
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   //
             (0<<TWWC);                         //
      break;

   case TWI_STX_DATA_NACK:          				// Data byte in TWDR has been transmitted; NACK has been received.
																// I.e. this could be the end of the transmission.

      TWI_state  = TWI_NO_STATE ;               // Das Auslesen ist beendet, die Empfangsbereitschaft wieder herstelle
      TWCR = (1<<TWEN)|                         // TWI Interface enabled.
            (1<<TWIE)|(1<<TWINT)|               // Enable TWI Interupt and clear the flag.
            (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|    // Prepare to ACK next time the Slave is addressed.
            (0<<TWWC);

      break;

	// ---  TWI Slave Receiver status codes ------------------------------------

//  case TWI_SRX_GEN_ACK:            				// General call address has been received; ACK has been returned
//  case TWI_SRX_GEN_ACK_M_ARB_LOST: 			   // Arbitration lost in SLA+R/W as Master;
																// General call address has been received; ACK has been returned

   case TWI_SRX_ADR_ACK:            				// Own SLA+W has been received ACK has been returned
//    case TWI_SRX_ADR_ACK_M_ARB_LOST: 			// Arbitration lost in SLA+R/W as Master;
																// own SLA+W has been received; ACK has been returned
                                                // Dont need to clear TWI_S_statusRegister.generalAddressCall
																// due to that it is the default state.

      TWI_Rx_buf[0] = TWDR;                     // Eigene Adresse an Pos. 0 zurckgeben - Kompatibilitt
		TWI_Rx_Idx = 1;                           // Set buffer pointer to first data location
                                                // Reset the TWI Interupt to wait for a new event.
      TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|              // Enable TWI Interupt and clear the flag to send byte
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Expect ACK on this transmission
             (0<<TWWC);                         //
      break;

   case TWI_SRX_ADR_DATA_ACK:       				// Previously addressed with own SLA+W;
																// data has been received; ACK has been returned
   case TWI_SRX_GEN_DATA_ACK:       				// Previously addressed with general call;
																// data has been received; ACK has been returned
		if (TWI_Rx_Idx < SLA_RX_BUF_SIZE)		   // berprfung des Pointers
			{
			TWI_Rx_buf[TWI_Rx_Idx++] = TWDR;	      // berzhlige Bytes werden nicht gespeichert !
			LastTransOK = TRUE;      					// Set flag transmission successfull.
                                                // Reset the TWI Interupt to wait for a new event.
			}
		// kein else: berzhlige Bytes werden ignoriert

		TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|              // Enable TWI Interupt and clear the flag to send byte
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Send ACK after next reception
             (0<<TWWC);                         //
      break;

   case TWI_SRX_STOP_RESTART:       				// A STOP condition or repeated START condition has been received
																// while still addressed as Slave
                                                // Put TWI Transceiver in passive mode.
      TWCR = (1<<TWEN)|                         // Enable TWI-interface and release TWI pins
             (0<<TWIE)|(1<<TWINT)|              // Disable Interupt
             (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Do not acknowledge on any new requests.
             (0<<TWWC);                         //
      break;


/* Der Rest luft ber default
   case TWI_SRX_ADR_DATA_NACK:      				// Previously addressed with own SLA+W;
																// data has been received; NOT ACK has been returned
   case TWI_SRX_GEN_DATA_NACK:      				// Previously addressed with general call;
																// data has been received; NOT ACK has been returned
   case TWI_STX_DATA_ACK_LAST_BYTE: 				// Last data byte in TWDR has been transmitted (TWEA = 0);
																// ACK has been received
   case TWI_NO_STATE:             				   // No relevant state information available; TWINT = 0
   case TWI_BUS_ERROR:         						// Bus error due to an illegal START or STOP condition
*/
   default:
      TWI_state = TWSR;                         // Store TWI State as errormessage, operation also clears the Success bit.
      TWCR = (1<<TWEN)|                         // Enable TWI-interface and release TWI pins
             (1<<TWIE)|(0<<TWINT)|              // Disable Interupt
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Do not acknowledge on any new requests.
             (0<<TWWC);                         //
	}
}

// --- [ eof ] ----------------------------------------------------------------
